(***********************************************************************
 *
 * ParserMath version 1.1
 *
 * Copyright (c) 2011, 2014 Michel Deslierres
 * All rights reserved. Tous droits rservs.
 *
 * Read full license in the file license.txt
 * Lire la licence au complet dans le fichier license.txt
 *
 ***********************************************************************
 *
 * Changements / Changes
 *
 * - version 1.1
 *     Added many functions found in Math.pas to handle built-in
 *     functions.
 *     Added IsZero, NearZero, SameNumber and Epsilon for approximate
 *     comparisons of reals.
 *
 ************************************************************************)

{: Unit that provides some mathematical functions not found in Math.pas
   and replacement functions if Math.pas is not available.}

unit ParserMath;

{$IFNDEF PARSER_INC}
  {$I Parser.inc}
{$ENDIF}

interface

uses
  ParserTypes;

var
  Epsilon: float;

function IsZero(const x: float): Boolean;
function NearZero(const x, eps: float): boolean;
function SameNumber(const x, y, eps: float): Boolean;

  {: Euclidean division, combined div and mod }
procedure DivModE(const X, Y: float; var Q: Integer; var R: float);

  {: Euclidean modulo division }
function ModE(const X, Y: float): float;

  {: Euclidean integer division }
function DivE(const X, Y: float): Integer;

  {: Truncated division, combined div and mod, Delphi compatible
     but defined for real operands }
procedure DivModT(const X, Y: float; var Q: Integer; var R: float);

  {: Truncated modulo division, Delphi compatible but defined
     for real operands }
function ModT(const X, Y: float): float;

  {: Truncated integer division, Delphi compatible but defined for
     real operands }
function DivT(const X, Y: float): Integer;

(************************************************************)
{$IFNDEF USE_MATH_UNIT}{-------------------------------------}
(************************************************************)

function Power(const Base, Exponent: float): float;

function Tan(const X: float): float;
function CoTan(const X: float): float;
function Secant(const X: float): float;
function Cosecant(const X: float): float;

function ArcSin(const X : float) : float;
function ArcCos(const X : float) : float;
function ArcCot(const X: float): float;
function ArcSec(const X: float): float;
function ArcCsc(const X: float): float;

function Cosh(const X: float): float;
function Sinh(const X: float): float;
function Tanh(const X: float): float;
function CotH(const X: float): float;
function SecH(const X: float): float;
function CscH(const X: float): float;

function ArcCosh(const X: float): float;
function ArcSinh(const X: float): float;
function ArcTanh(const X: float): float;
function ArcCotH(const X: float): float;
function ArcSecH(const X: float): float;
function ArcCscH(const X: float): float;

function LogN(const Base, X: float): float;
function Log10(const X: float): float;

function DegToRad(const Degrees: float): float;
function RadToDeg(const Radians: float): float;

function Ceil(const X: float):Integer;
function Floor(const X: float): Integer;

{$ENDIF}

implementation


(***************************************************
Reference:

Division and Modulus for Computer Scientists
DAAN LEIJEN
University of Utrecht
Dept. of Computer Science
PO.Box 80.089, 3508 TB Utrecht
The Netherlands
daan@cs.uu.nl,http://www.cs.uu.nl/~daan/lvm.html
December 3, 2001

/* Euclidean division */
long divE( long D, long d )
{
  long q = D/d;
  long r = D%d;
  if (r < 0) {
    if (d > 0) q = q-1;
          else q = q+1;
  }
  return q;
}

long modE( long D, long d )
{
  long r = D%d;
  if (r < 0) {
    if (d > 0) r = r + d;
          else r = r - d;
 }
 return r;
}

*******************************************************)

procedure DivModE(const X, Y: float; var Q: Integer; var R: float);
begin
  Q := trunc(X/Y);
  R := X - Q*Y;
  if R < 0 then begin
    if Y > 0 then begin
      R := R + Y;
      dec(Q)
    end
    else begin
      R := R - Y;
      Inc(Q);
    end;
  end;
end;

function ModE(const X, Y: float): float;
var
  Q: integer;
begin
  //  DivModE(X,Y,Q, result);
  Q := trunc(X/Y);
  result := X - Q*Y;
  if result < 0 then begin
    if Y > 0 then begin
      result := result + Y;
    end
    else begin
      result := result - Y;
    end;
  end;
end;

function DivE(const X, Y: float): Integer;
begin
  //DivModE(X,Y,result, R);
  result := trunc(X/Y);
  if X - Result*Y < 0 then begin
    if Y > 0 then begin
      dec(result)
    end
    else begin
      Inc(result);
    end;
  end;
end;

procedure DivModT(const X, Y: float; var Q: Integer; var R: float);
begin
  Q := trunc(X/Y);
  R := X - Q*Y
end;

function ModT(const X, Y: float): float;
begin
  result := X - trunc(X/Y)*Y;
end;

function DivT(const X, Y: float): Integer;
begin
  result := trunc(X/Y)
end;

function IsZero(const x: float): Boolean;
begin
  Result := Abs(x) <= Epsilon;
end;

function NearZero(const X, eps: float): boolean;
var
 ep: float;
begin
  if Eps <= 0 then
    ep := Epsilon
  else
    ep := eps;
  result := abs(X) <= Ep;
end;

function SameNumber(const x, y, eps: float): Boolean;
var
  a, b, ep: float;
begin
  if Eps = 0 then
    ep := Epsilon
  else
    ep := eps;
  if (y <> 0) and (x <> 0) then begin
    a := abs(x);
    b := abs(y);
    if a > b then
      a := b;
    ep := a*ep
  end;
  Result := abs(x - y) <= ep;
end;

procedure CalcMachineEpsilon;
var
  One: float;
  T: float;
begin
  One := 1.0;
  Epsilon := One;
  repeat
    Epsilon := 0.5 * Epsilon;
    T := One + Epsilon;
  until One = T;
  Epsilon := 2.0 * Epsilon;
end;

(************************************************************)
{$IFNDEF USE_MATH_UNIT}{-------------------------------------}
(************************************************************)

function Power(const Base, Exponent: float): float;
begin
  if Exponent = 0.0 then
    Result := 1.0     // 0**0 = 1
  else
    Result := Exp(Exponent * Ln(Base))
end;

function Tan(const X: float): float;
begin
  result := sin(x)/cos(x)
end;

function CoTan(const X: float): float;
begin
  result := 1/Tan(X)
end;

function Secant(const X: float): float;
begin
  result := 1 / Cos(X)
end;

function Cosecant(const X: float): float;
begin
  result := 1 / Sin(X)
end;

function arctan2(const y,x : float): float;
begin
  if x <> 0 then
    result := ArcTan(y/x)
  else begin
    if y = 0 then
      result := 0.0
    else if y > 0 then
      result := pi/2
    else if y<0 then
      result := -pi/2;
  end;
  if x < 0.0 then
    result := result + pi;
  if result > pi then
    result := result - 2*pi;
end;

function ArcSin(const X : float) : float;
begin
  Result := ArcTan2(X, Sqrt((1+X) * (1-X)))
end;

function ArcCos(const X : float) : float;
begin
  Result := ArcTan2(Sqrt((1+X) * (1-X)), X)
end;

function Cosh(const X: float): float;
begin
  if IsZero(X) then
    Result := 1
  else
    Result := (Exp(X) + Exp(-X)) / 2;
end;

function Sinh(const X: float): float;
begin
  if IsZero(X) then
    Result := 0
  else
    Result := (Exp(X) - Exp(-X)) / 2;
end;

function Tanh(const X: float): float;
begin
  if IsZero(X) then
    Result := 0
  else
    Result := SinH(X) / CosH(X);
end;

function ArcCosh(const X: float): float;
begin
  Result := Ln(X + Sqrt((X - 1) / (X + 1)) * (X + 1));
end;

function ArcSinh(const X: float): float;
begin
  Result := Ln(X + Sqrt((X * X) + 1));
end;

function ArcTanh(const X: float): float;
begin
  Result := 0.5 * Ln((1 + X) / (1 - X));
end;

function CotH(const X: float): float;
begin
  Result := 1 / TanH(X);
end;

function SecH(const X: float): float;
begin
  Result := 1 / CosH(X);
end;

function CscH(const X: float): float;
begin
  Result := 1 / SinH(X);
end;

function ArcCot(const X: float): float;
begin
  if IsZero(X) then
    Result := PI / 2
  else
    Result := ArcTan(1 / X);
end;

function ArcSec(const X: float): float;
begin
  Result := ArcCos(1 / X);
end;

function ArcCsc(const X: float): float;
begin
  Result := ArcSin(1 / X);
end;

function ArcCotH(const X: float): float;
begin
  Result := 0.5 * Ln((X + 1) / (X - 1));
end;

function ArcSecH(const X: float): float;
begin
  Result := Ln((Sqrt(1 - X * X) + 1) / X);
end;

function ArcCscH(const X: float): float;
begin
  Result := Ln(Sqrt(1 + (1 / (X * X))) + (1 / X));
end;

function LogN(const Base, X: float): float;
begin
  result := ln(X) / ln(Base);
end;

function Log10(const X: float): float;
begin
  result := ln(X) / ln(10);
end;

function DegToRad(const Degrees: float): float;  { Radians := Degrees * PI / 180 }
begin
  Result := Degrees * (PI / 180);
end;

function RadToDeg(const Radians: float): float;  { Degrees := Radians * 180 / PI }
begin
  Result := Radians * (180 / PI);
end;

function Ceil(const X: float): Integer;
begin
  Result := Integer(Trunc(X));
  if Frac(X) > 0 then
    Inc(Result);
end;

function Floor(const X: float): Integer;
begin
  Result := Integer(Trunc(X));
  if Frac(X) < 0 then
    Dec(Result);
end;

{$ENDIF}

initialization
  CalcMachineEpsilon;
end.
