(***********************************************************************
 *
 * ParserTypes version 1.3
 *
 * Copyright (c) 2013, 2014 Michel Deslierres
 * All rights reserved. Tous droits rservs.
 *
 * Read full license in the file license.txt
 * Lire la licence au complet dans le fichier license.txt
 *
 ***********************************************************************
 *
 * Changements / Changes
 *
 * - version 1.3
 *     Add peUnknownError to the enumerated type TParseError.
 *
 * - version 1.2
 *     Added TTree the abstract class for all nodes of a parse tree.
 *
 * - version 1.1
 *     Modified to handle internal functions by adding
 *     types pfloat, floatArray, pFloatArray and parser errors
 *     peExpectedSeparator, peExpectedLeftPar, and
 *     peUnknownIdentifier.
 *     Moved EMathParserErr here.
 *
 ************************************************************************)

{: Unit that contains global types that are used by other units in
   the project. See ParserConsts.pas for string constants.}
Unit ParserTypes;

{$IFNDEF PARSER_INC}
  {$I Parser.inc}
{$ENDIF}


interface

Uses
{$IFDEF DUMP}
  ComCtrls,
{$ENDIF}
  SysUtils;

type
  float = Extended;
  pfloat = ^float;
  floatArray = array[0..maxint div sizeof(float)-1] of float;
  pFloatArray = ^floatArray;

  {: Abstract ancestor class of all classes created by the parser,
     see Parse in TreeMathParser}
  TTree = class
  public
    {: The function that evaluates a tree. Can raise an exception of type
       EMathError (EInvalidOp, EZeroDivide, EOverflow and EUnderflow)}
    function Eval: float; virtual; abstract;

    {$IFDEF DUMP}
    function ToText: string; virtual;
    procedure ToTreeView(tv: TTreeView);
    {$ENDIF}
  end;

  {: Errors that the mathematical expression parser may report.}
  TParserError = (
    {: Never returned by parser}
    peNoError,

    {: Unexpected end of string when expecting a factor. This error
       will be raised when parsing "5 +" because a factor should
       follow the + sign.}
    peUnexpectedEOS,

    {: Expected the end of the expression, found something extraneous.
       This error occurs when parsing the final number in the expression
       "5 + 8 32".}
    peExpectedEOS,

    {: An invalid number was found. This occurs in the Factor routine
       when parsing an expression such as "5,3E*8" where the "E" must
       be followed by a valid positive or negative integer.}
    peInvalidNumber,

    {: Found an unknown or invalid operator when parsing a factor.
       This errors occurs when parsing the multiply operator in the
       expression "5 + *" or when parsing the unknown operator "!"
       in the expression "5 + !".}
    peUnexpectedElement,

    {: Expected a ")" to mark the end of a function call or to mark the
       end of an expression that started with a "(". This error occurs
       in the procedure Builtin when parsing the expressions "sin(5 + 8"
       and "sin(5, 8)"  and in the procedure Factor when parsing a factor
       of the form "( expression )" with a missing right paranthesis as
       in the expression "5*(2 + 3".}
    peExpectedRightPar,

    {: Expected list separator (see ParserListChar and SetParserListChar)
       when parsing a function's arguments. This error will occur when
       parsing  "max(4 5)" or "max(4 / 5)".}
    peExpectedSeparator,

    {: Expected a "(" after an function identifier. This error will occur
       when parsing "ln 32".}
    peExpectedLeftPar,

    {: An unknown identifier has been found. Example "lin(32)"
       assuming no built in nor user defined function of that name
       exists.}
    peUnknownIdentifier,

    {: Too few parameters were found when parsing a function's
       parameter list. Example "min()"}
    peTooFewParams,

    {: Too many parameters were found when parsing a function's
       parameter list. Example "sin(32, 34)"}
    peTooManyParams,

    {: A user definition consists of a name, optionnally followed by a
       parameter list, then the assignment symbol ":=" and finally
       an expression.}
    peExpectedIdentifier,

    {: A user definition must have a unique name.}
    peDuplicateIdentifier,

    {$IFNDEF CAN_REDEFINE_BUILTIN_FUNCTION}
    {: A user definition's name cannot be the same as a buitin's name.}
    peIdentIsBuiltInName,
    {$ENDIF}

    {: A user definition consists of a name, optionnally followed by a
       parameter list, and then the assignment symbol ":=".}
    peExpectedAssign,

    {: A user definition's formal parameter list is missing an identifier.
       Can occur if the source is
         a) f()
         b) f(a, )
         c) f(43)
         .}
    peExpectedFormalParameter,

    {: A formal parameter has the same name as the user definition's name.
       For example, f(a, f) := a*f  which would be difficult to parse correctly.
       In the expression a*f does f refer to the second formal parameter or
       the function.}
    peInvalidFormalParameterName,

    {: A formal parameter has the same name as another formal parameter in
       a user definition. For example f(x, x) := 5*x which would be ambiguous.}
    peDuplicateFormalParameterName,

    {: f(x 32 y) }
    peExpectedSeparatorOrRightPar,

    {: Error never returned by the parser. It was added to provide a generic
       error message for other execptions because parsing assign statements
       will not raise exceptions}
    peUnknownError
    );

  {: Parser exception }
  EMathParserErr = class(Exception)
  public
    ParserError: TParserError;
    TokenStart: integer;
    TokenLength: integer;
  end;

implementation


{$IFDEF DUMP}

uses
  Classes;

function TTree.ToText: string;
var
  s: string;
begin
  try
    s := Format('%g', [eval])
  except
    s := 'nan'
  end;
  result := Format(' (eval = %s)', [s]);
end;

procedure TTree.ToTreeView(tv: TTreeView);
{$IFDEF HAS_TSTRINGSTREAM}
var
  s: TStringStream;
begin
  tv.Items.Clear;
  s := TStringStream.Create(ToText);
  try
{$ELSE}
var
  size: integer;
  {$IFDEF UNICODE}
  t: ansistring;
  {$ELSE}
  t: string;
  {$ENDIF}
  s: TMemoryStream;
begin
  t := ToText;// + #0#0;
  size := length(t);
  s := TMemoryStream.Create;
  try
    s.setsize(size);
    s.WriteBuffer(pointer(t)^, size);
{$ENDIF}
    s.Position := 0;
    tv.LoadFromStream(s);
    tv.FullExpand;
  finally
    s.Free;
  end;
end;
{$ENDIF}

end.


