(***********************************************************************
 *
 * Targets 1.2
 *
 * Copyright (c) 2014 Michel Deslierres
 * All rights reserved. Tous droits rservs.
 *
 * Read full license in the file license.txt
 * Lire la licence au complet dans le fichier license.txt
 *
 ************************************************************************
 *
 * Changes/Changement
 *
 * - version 1.2
 *     Added if, near and ins and outs functions.
 *
 * - version 1.1
 *     Added two other types of targets: variables and user defined
 *     functions. The latter are added to a different list. This 
 *     allows the user to redefine built-in functions although in the
 *     present version it is not allowed.
 *
 ************************************************************************
 *  A TCallNode defined by the parser can now call on six types of
 *  targets:
 *     - built-in constant (pi, e)
 *     - built-in function with fixed number of arguments (sin, log, ...)
 *     - built-in function with variable number of arguments (max, min, ...)
 *     - built-in variable
 *     - user defined constant
 *     - user defined function with fixed number of argmuments
 ************************************************************************)

 {: Unit that defines the list of internal (built-in) and external (user-
   defined) functions and constants.}
unit Targets;

{$IFNDEF PARSER_INC}
  {$I Parser.inc}
{$ENDIF}

interface

uses
   Classes, ParserTypes;

Type
  {: Base class of all named targets. An individual target does not know
     its own name.}
  TTarget = class
  private
    FCount: integer;
  protected
	  procedure SetCount(value: integer); virtual;
  public
    function Calc: float; virtual; abstract;
    property Count: integer read FCount write SetCount;
  end;

  {: Class of built-in constants. Create with the constant's value
     as in TConstant.Create(pi). }
  TConstant = class(TTarget)
  private
    FValue: float;
  public
    constructor create(value: float);
    function Calc: float; override;
  end;

  {: Class of built-in variables. Created with the address of variable
     as in
       var
         x: float;

       TVariable.Create(@x) }
  TVariable = class(TTarget)
  private
    FValue: pfloat;
  public
    constructor create(value: pfloat);
	  function Calc: float; override;
  end;

  {: Class of named functions with fixed number of parameters.
     Created with number of variables which should be 1 or greater.
     Note that the function does not own the Args array. The call node
     must set the function's ArgArray to its array of args before
     calling Calc.

     Calculation of a built-in function is done in two steps. First
     the array of varlues, one for each argument of the funciton
     is calculated by the caller, the built-in function's ArgArray
     is set equal to that array and then the built-in function's
     Calc function is called.

        Func.ArgArray := Args;
        result := Func.Calc

    See Call in the parser.

    The class TFunction does not allocate memory to its FArgs field,
    it does not free any memory to that pointer. This is the 
    responsability of the caller. }
  TFunction = class(TTarget)
  private
    FArgs: pFloatArray;
    function GetArg(index: integer): float;
  public
    constructor create(aCount: integer);
    property Args[index: integer]: float read GetArg;
    property ArgArray: pFloatArray read FArgs write FArgs;
  end;

  {: Class of named functions with variable number of parameters, create with
     the minimum number of expected variables.}
  TVariableFunction = class(TFunction)
  private
    FMinCount: integer;
  protected
    procedure SetCount(value: integer); override;
  public
    constructor create(aMinCount: integer);
    property MinCount: integer read FMinCount;
  end;

  {: User functions are defined from an assigment statement
       f(x, y) := 2*x + y
     A parse tree of the right hand side is constructed
     and assigned to the target's Tree property.}
  TUserFunction = class(TFunction)
  private
    FTree: TTree;
  public
    constructor create(aTree: TTree; ParamCount: integer);
    destructor Destroy; override;
	  function Calc: float; override;
    property Tree: TTree read FTree write FTree;
  end;

  TTargetList = class
  private
    FList: TStrings;
  public
    constructor Create;
    destructor Destroy; override;

    {: Add an internal target by supplying it's name and instantiated class.
       If an error occurs (duplicate name) then target is freed and
       an exception is raised. }
    procedure Add(const aName: string; target: TTarget);

	  {: Removes all internal targets from the list }
    procedure Clear;

	  {: Removes a named internal target from the list. Returns true
       if aName was found and it's target freed, false otherwise. }
    function Delete(const aName: string): boolean;

	  {: Searches for a target by name. If found, returns true and returns
       the found target.}
    function Find(const aName: string; var target: TTarget): boolean;

    {: Returns true if the specified name is found in the list. }
    function HasTarget(const aName: string): boolean;

    {$IFDEF DUMP}
	  {: Returns the name of function, constant or variable that has the
	     specified evaluator.}
    function NameOf(target: TTarget): string;
    {$ENDIF}
  end;


var
  {Built-in and user-defined functions are kept in separate lists.
   When an identifier is found, ExternalTargets is searched first
   which would allow redefining built-in functions by users.}
  InternalTargets: TTargetList;
  ExternalTargets: TTargetList;

implementation

uses
  SysUtils,
  {$IFDEF HAS_SYSCONST} SysConst, {$ENDIF}
  ParserConsts, ParserMath {$IFDEF USE_MATH_UNIT}, Math {$ENDIF};

procedure TTarget.SetCount(value: integer);
begin
  {$IFDEF DEBUG}
  assert(false, 'Cannot set Count');
  {$ENDIF}
end;

constructor TConstant.create(value: Extended);
begin
  inherited Create;
  //FCount := 0;
  FValue := value;
end;

function TConstant.Calc: float;
begin
  result := FValue;
end;

constructor TVariable.create(value: pfloat);
begin
  inherited Create;
  //FCount := 0;
  FValue := value;
end;

function TVariable.Calc: float;
begin
  result := FValue^;
end;

{ TFunction }

constructor TFunction.create(aCount: integer);
begin
  inherited Create;
  FCount := aCount;
end;

function TFunction.GetArg(index: integer): float;
begin
  {$IFDEF DEBUG}
  if (index < 0) or (index >= FCount) then
    raise ERangeError.CreateRes(@SRangeError);
  {$ENDIF}
  result := FArgs^[index]
end;

constructor TVariableFunction.Create(aMinCount: integer);
begin
  inherited Create(0);  //FCount := 0
  {$IFDEF DEBUG}
  assert(aMinCount > 0, 'Minimum number of arguments less than 1');
  {$ENDIF}
  FMinCount := aMinCount;
end;

procedure TVariableFunction.SetCount(value: integer);
begin
  if Count <> value then begin
    {$IFDEF DEBUG}
    assert(value >= MinCount, 'Too few parameters');
    {$ENDIF}
    FCount := value;
  end;
end;

{ TUserFunction }

constructor TUserFunction.create(aTree: TTree; ParamCount: integer);
begin
  inherited Create(ParamCount);
  FTree := aTree;
end;

destructor TUserFunction.destroy;
begin
  FTree.Free;
  inherited;
end;

function TUserFunction.Calc: float;
begin
  result := FTree.Eval;
end;

{ TTargetList }

procedure TTargetList.add(const aName: string; target: TTarget);
begin
  try
    FList.AddObject(aName, target);
  except
    target.Free;
    Raise;
  end;
end;

procedure TTargetList.clear;
var
  i: integer;
begin
  TStringList(FList).BeginUpdate;
  try
    for i := 0 to FList.Count - 1 do
      FList.Objects[i].Free;
    FList.Clear;
  finally
    TStringList(FList).EndUpdate;
  end;
end;

constructor TTargetList.create;
begin
  inherited;
  FList := TStringList.create;
  TStringList(FList).Duplicates := dupError;
  TStringList(FList).Sorted := true;
end;

function TTargetList.delete(const aName: string): boolean;
var
  i: integer;
begin
  result := TStringList(FList).Find(aName, i);
  if result then begin
    FList.Objects[i].Free;
    FList.Delete(i);
  end;
end;

destructor TTargetList.destroy;
begin
  Clear;
  FList.free;
  inherited;
end;

function TTargetList.find(const aName: string; var target: TTarget): boolean;
var
  i: integer;
begin
  result := TStringList(FList).Find(aName, i);
  if result then begin
    target := FList.Objects[i] as TTarget;
  end
  else begin
    target := nil;
  end;
end;

function TTargetList.HasTarget(const aName: string): boolean;
var
  i: integer;
begin
  result := TStringList(FList).Find(aName, i)
end;

{$IFDEF DUMP}
type
  TFunctionX = record
    argcount: integer;
    case boolean of
      //false: (calc: CalcProc);
      true: (calcp: pointer);
    end;
  PFunctionx = ^TFunctionx;

function TTargetList.NameOf(target: TTarget): string;
var
  i: integer;
begin
  for I := 0 to FList.Count - 1 do
    if FList.Objects[i] as TTarget = target then begin
      result := FList[i];
      exit;
    end;
  result := _(SUnknown);
end;
{$ENDIF}

{trigonometric}

type
  _sin = class(TFunction)
  public
    function Calc: float; override;
  end;

  _cos = class(TFunction)
  public
    function Calc: float; override;
  end;

  _tan = class(TFunction)
  public
    function Calc: float; override;
  end;

  _cot = class(TFunction)
  public
    function Calc: float; override;
  end;

  _csc = class(TFunction)
  public
    function Calc: float; override;
  end;

  _sec = class(TFunction)
  public
    function Calc: float; override;
  end;

  _rad = class(TFunction)
  public
    function Calc: float; override;
  end;

  _deg = class(TFunction)
  public
    function Calc: float; override;
  end;

function _sin.Calc: float;
begin
  result := sin(Args[0]);
end;

function _cos.Calc: float;
begin
  result := cos(Args[0]);
end;

function _tan.Calc: float;
begin
  result := tan(Args[0]);
end;

function _cot.Calc: float;
begin
  result := cotan(Args[0]);
end;

function _csc.Calc: float;
begin
  result := cosecant(Args[0]);
end;

function _sec.Calc: float;
begin
  result := secant(Args[0]);
end;

function _rad.Calc: float;
begin
  result := DegToRad(Args[0]);
end;

function _deg.Calc: float;
begin
  result := RadToDeg(Args[0]);
end;

{inverse trigonometric}

type
  _asin = class(TFunction)
  public
    function Calc: float; override;
  end;

  _acos = class(TFunction)
  public
    function Calc: float; override;
  end;

  _atan = class(TFunction)
  public
    function Calc: float; override;
  end;

  _acot = class(TFunction)
  public
    function Calc: float; override;
  end;

  _acsc = class(TFunction)
  public
    function Calc: float; override;
  end;

  _asec = class(TFunction)
  public
    function Calc: float; override;
  end;

function _asin.Calc: float;
begin
  result := arcsin(Args[0]);
end;

function _acos.Calc: float;
begin
  result := arccos(Args[0]);
end;

function _atan.Calc: float;
begin
  result := arctan(Args[0]);
end;

function _acot.Calc: float;
begin
  result := arctan(1/Args[0]);
end;

function _acsc.Calc: float;
begin
  result := arcsin(1/Args[0]);
end;

function _asec.Calc: float;
begin
  result := arccos(1/Args[0]);
end;


{hyperbolic}

type
  _sinh = class(TFunction)
  public
    function Calc: float; override;
  end;

  _cosh = class(TFunction)
  public
    function Calc: float; override;
  end;

  _tanh = class(TFunction)
  public
    function Calc: float; override;
  end;

  _coth = class(TFunction)
  public
    function Calc: float; override;
  end;

  _csch = class(TFunction)
  public
    function Calc: float; override;
  end;

  _sech = class(TFunction)
  public
    function Calc: float; override;
  end;

function _sinh.Calc: float;
begin
  result := sinh(Args[0]);
end;

function _cosh.Calc: float;
begin
  result := cosh(Args[0]);
end;

function _tanh.Calc: float;
begin
  result := tanh(Args[0]);
end;

function _coth.Calc: float;
begin
  result := coth(Args[0]);
end;

function _csch.Calc: float;
begin
  result := csch(Args[0]);
end;

function _sech.Calc: float;
begin
  result := sech(Args[0]);
end;

{inverse hyperbolic}

type
  _asinh = class(TFunction)
  public
    function Calc: float; override;
  end;

  _acosh = class(TFunction)
  public
    function Calc: float; override;
  end;

  _atanh = class(TFunction)
  public
    function Calc: float; override;
  end;

  _acoth = class(TFunction)
  public
    function Calc: float; override;
  end;

  _acsch = class(TFunction)
  public
    function Calc: float; override;
  end;

  _asech = class(TFunction)
  public
    function Calc: float; override;
  end;

function _asinh.Calc: float;
begin
  result := arcsinh(Args[0]);
end;

function _acosh.Calc: float;
begin
  result := arccosh(Args[0]);
end;

function _atanh.Calc: float;
begin
  result := arctanh(Args[0]);
end;

function _acoth.Calc: float;
begin
  result := arccoth(Args[0]);
end;

{ArcCscH is incorrectly implemented in Math.pas
 included in versions of Delphi prior to Rad Studio 8.0
 (Delphi XE).}

{$IFDEF FIX_ARCCSCH}
  {$IFNDEF USE_MATH_UNIT}
    should not be here!
  {$ENDIF}
function _acsch.Calc: float;
begin
  Result := Ln(Sqrt(1 + (1 / (Args[0] * Args[0]))) + (1 / Args[0]));
end;
{$ELSE}
function _acsch.Calc: float;
begin
  result := arccsch(Args[0])
end;
{$ENDIF}

function _asech.Calc: float;
begin
  result := arcsech(Args[0]);
end;

{general}

type
  _abs = class(TFunction)
  public
    function Calc: float; override;
  end;

  _sqr = class(TFunction)
  public
    function Calc: float; override;
  end;

  _sqrt = class(TFunction)
  public
    function Calc: float; override;
  end;

  _ln = class(TFunction)
  public
    function Calc: float; override;
  end;

  _log = class(TFunction)
  public
    function Calc: float; override;
  end;

  _log10 = class(TFunction)
  public
    function Calc: float; override;
  end;

  _exp = class(TFunction)
  public
    function Calc: float; override;
  end;

  _odd = class(TFunction)
  public
    function Calc: float; override;
  end;

  _floor = class(TFunction)
  public
    function Calc: float; override;
  end;

  _ceil = class(TFunction)
  public
    function Calc: float; override;
  end;

  _ins = class(TFunction)
  public
    function Calc: float; override;
  end;

  _int = class(TFunction)
  public
    function Calc: float; override;
  end;

  _outs = class(TFunction)
  public
    function Calc: float; override;
  end;

  _sgn = class(TFunction)
  public
    function Calc: float; override;
  end;

  _round = class(TFunction)
  public
    function Calc: float; override;
  end;

  _clamp = class(TFunction)
  public
    function Calc: float; override;
  end;

  _near = class(TFunction)
  public
    function Calc: float; override;
  end;

  _fact = class(TFunction)
  public
    function Calc: float; override;
  end;

  _max = class(TVariableFunction)
  public
    function Calc: float; override;
  end;

  _min = class(TVariableFunction)
  public
    function Calc: float; override;
  end;

  _avg = class(TVariableFunction)
  public
    function Calc: float; override;
  end;

  _ssq = class(TVariableFunction)
  public
    function Calc: float; override;
  end;
  
  _var = class(TVariableFunction)
  public
    function Calc: float; override;
  end;

  _varp = class(TVariableFunction)
  public
    function Calc: float; override;
  end;
  
  _poly = class(TVariableFunction)
  public
    function Calc: float; override;
  end;
 
function _abs.Calc: float;
begin
  result := abs(Args[0]);
end;

function _sqr.Calc: float;
begin
  result := sqr(Args[0]);
end;

function _sqrt.Calc: float;
begin
  result := sqrt(Args[0]);
end;

function _ln.Calc: float;
begin
  result := ln(Args[0]);
end;

function _log.Calc: float;
begin
  result := logn(Args[1], Args[0]);
end;

function _log10.Calc: float;
begin
  result := log10(Args[0]);
end;

function _exp.Calc: float;
begin
  result := exp(Args[0]);
end;

function _odd.Calc: float;
begin
  if odd(trunc(Args[0])) then
    result := 1
  else
    result := 0;
end;

function _floor.Calc: float;
begin
  result := Floor(Args[0]);
end;

function _ceil.Calc: float;
begin
  result := Ceil(Args[0]);
end;

function _ins.Calc: float;
begin
  result := 1.0;
  if args[1] < args[2] then begin
    if (args[0] < args[1])
    or (args[0] > args[2]) then
      result := 0.0
  end
end;

function _int.Calc: float;
begin
  result := trunc(Args[0]);
end;

function _outs.Calc: float;
begin
  result := 1.0;
  if args[1] < args[2] then begin
    if (args[1] <= args[0])
    and (args[0] <= args[2]) then
      result := 0.0
  end
end;

function _sgn.Calc: float;
begin
  if args[0] > 0 then
    result := 1
  else if args[0] < 0 then
    result := -1
  else
    result := 0;
end;

function _round.Calc: float;
begin
  result := round(Args[0]);
end;

function _clamp.Calc: float;
begin
  result := args[0];
  if args[1] <= args[2] then begin
    if result <  args[1]then
      result := args[1]
    else if result > args[2] then
      result := args[2]
  end;
end;

function _near.Calc: float;
begin
  if SameNumber(args[0], args[1], args[2]) then
    result := 1
  else
    result := 0;
end;

function _fact.Calc: float;
var
  x: float;
begin
  if args[0] < 0 then
    result := ln(-1)
  else begin
    x := trunc(args[0]);
    result := 1;
    while x > 1 do begin
      result := result * x;
      x := x - 1;
    end;
  end;
end;

{ variable functions }

function _max.Calc: float;
var
  i: integer;
begin
  result := Args[0];
  for i := 1 to Count-1 do
    if Args[i] > result then
      result := Args[i]
end;

function _min.Calc: float;
var
  i: integer;
begin
  result := Args[0];
  for i := 1 to Count-1 do
    if Args[i] < result then
      result := Args[i]
end;

function _avg.Calc: float;
var
  i: integer;
begin
 result := 0;
  for i := 0 to Count-1 do
    result := result + args[i]/count;
end;

function _ssq.Calc: float;
var
  i: integer;
begin
  result := 0;
  for i := 0 to Count-1 do
    result := result + sqr(args[i]);
end;

function _var.Calc: float;
var
  i,n: integer;
  avg: float;
begin
  avg := 0;
  for i := 0 to Count-1 do
    avg := avg + args[i]/count;
  n := pred(count);
  result := 0;
  for i := 0 to Count-1 do
    result := result + (sqr(args[i]-avg))/n;
end;

function _varp.Calc: float;
var
  i: integer;
  avg: float;
begin
  avg := 0;
  for i := 0 to Count-1 do
    avg := avg + args[i]/count;
  result := 0;
  for i := 0 to Count-1 do
    result := result + (sqr(args[i]-avg))/count;
end;

function _poly.Calc: float;
// x, a0, a1, a2
// 0   1   2   3   4
// computes a0 + a1*x + a2*x^2 + a3*x^3... a(count-1)*x^(count-1)
var
  i: integer;
  x: float;
begin
  x := args[0];
  result := 0;
  for i := Count-1 downto 1 do
    result := args[i] + x*result;
end;

procedure LoadInternalTargets;
begin
  InternalTargets.Add('sin', _sin.create(1));
  InternalTargets.Add('cos', _cos.create(1));
  InternalTargets.Add('tan', _tan.create(1));
  InternalTargets.Add('cot', _cot.create(1));
  InternalTargets.Add('csc', _csc.create(1));
  InternalTargets.Add('sec', _sec.create(1));
  InternalTargets.Add('rad', _rad.create(1));
  InternalTargets.Add('deg', _deg.create(1));

  InternalTargets.Add('asin', _asin.create(1));
  InternalTargets.Add('acos', _acos.create(1));
  InternalTargets.Add('atan', _atan.create(1));
  InternalTargets.Add('acot', _acot.create(1));
  InternalTargets.Add('acsc', _acsc.create(1));
  InternalTargets.Add('asec', _asec.create(1));

  InternalTargets.Add('sinh', _sinh.create(1));
  InternalTargets.Add('cosh', _cosh.create(1));
  InternalTargets.Add('tanh', _tanh.create(1));
  InternalTargets.Add('coth', _coth.create(1));
  InternalTargets.Add('csch', _csch.create(1));
  InternalTargets.Add('sech', _sech.create(1));

  InternalTargets.Add('asinh', _asinh.create(1));
  InternalTargets.Add('acosh', _acosh.create(1));
  InternalTargets.Add('atanh', _atanh.create(1));
  InternalTargets.Add('acoth', _acoth.create(1));
  InternalTargets.Add('acsch', _acsch.create(1));
  InternalTargets.Add('asech', _asech.create(1));

  InternalTargets.Add('abs', _abs.create(1));
  InternalTargets.Add('sqr', _sqr.create(1));
  InternalTargets.Add('sqrt', _sqrt.create(1));
  InternalTargets.Add('ln', _ln.create(1));
  InternalTargets.Add('log', _log.create(2));
  InternalTargets.Add('log10', _log10.create(1));
  InternalTargets.Add('exp', _exp.create(1));
  InternalTargets.Add('odd', _odd.create(1));
  InternalTargets.Add('floor', _floor.create(1));
  InternalTargets.Add('ceil', _ceil.create(1));
  InternalTargets.Add('int', _int.create(1));
  InternalTargets.Add('sgn', _sgn.create(1));
  InternalTargets.Add('round', _round.create(1));
  InternalTargets.Add('fact', _fact.create(1));
  InternalTargets.Add('ins', _ins.create(3));
  InternalTargets.Add('outs', _outs.create(3));
  InternalTargets.Add('clamp', _clamp.create(3));
  InternalTargets.Add('near', _near.create(3));

  InternalTargets.Add('max', _max.create(1));
  InternalTargets.Add('min', _min.create(1));
  InternalTargets.Add('avg', _avg.create(1));
  InternalTargets.Add('ssq', _ssq.create(1));
  InternalTargets.Add('var', _var.create(2));
  InternalTargets.Add('varp', _varp.create(1));
  InternalTargets.Add('poly', _poly.create(2));

  InternalTargets.Add('e', TConstant.create(exp(1)));
  InternalTargets.Add('pi', TConstant.create(pi));
  
  InternalTargets.Add('if', nil);
end;

initialization
  ExternalTargets := TTargetList.Create;
  InternalTargets := TTargetList.create;
  LoadInternalTargets;
finalization
  InternalTargets.Free;
  ExternalTargets.Free;
end.
