unit TestParser;

interface

{$I ..\Parser.inc}

uses
  Math, TestFramework, SysUtils, Classes, ParserTypes, ParserConsts,
  Targets;

type
  TTestParserHarness = class(TTestCase)
  strict private
    FSource: string;
    oldExceptionMask: TFPUExceptionMask;
  protected
    procedure TestOk(const value: string; wanted: float; eps: float = 0; const msg: string = '');
    procedure TestBadParse(const value: string; error: TParserError);
    procedure TestBadEval(const value: string; error: ExceptClass);
  public
    procedure SetUp; override;
    procedure TearDown; override;
  end;

  TTestTreeParser = class(TTestParserHarness)
  published
    procedure TestCreate;
    procedure TestFactor0;
    procedure TestFactor10;
    procedure TestFactorMinus10;
    procedure TestFactorFloat;
    procedure TestInvalidNumber;
    procedure TestTerm1;
    procedure TestTerm2;
    procedure TestExpression1;
    procedure TestExpression2;
    procedure TestExpression3;
    procedure TestExpression4;
    procedure TestExpression5;
    procedure TestExpression6;
    procedure TestMultiline;
    procedure TestMultiline2;
    procedure TestUnexpectedEOS;
    procedure TestExpectedEOS;
    procedure TestUnexpectedElement;
    procedure TestExpecteRightPar;
    procedure TestDivBy0;
    procedure TestOverflow;
    procedure TestUnderflow;
    procedure TestIntDiv;
    procedure TestMod;
    procedure TestPower;
    procedure TestPower2;
    procedure TestPower3;
    procedure TestPower4;
    procedure TestFunctions;
    procedure TestWrongParamCount;
    procedure TestLess;
    procedure TestLessOrEqual;
    procedure TestGreater;
    procedure TestGreaterOrEqual;
    procedure TestEqual;
    procedure TestNotEqual;
    procedure TestIf;
    procedure TestOr;
    procedure TestAnd;
    procedure TestNot;
  end;

  TTestTargets = class(TTestParserHarness)
  published
    procedure TestAddConstant;
    procedure TestAddVariable;
    procedure TestAddFixedFunction;
    procedure TestAddVariableFunction;
  end;

implementation

uses
  MathParser;

procedure TTestParserHarness.SetUp;
// somewhere (?) exUnderflow is being turned off!
// FPUException = (exInvalidOp, exDenormalized, exZeroDivide,
//                   exOverflow, exUnderflow, exPrecision);
// Removing a TFpuException from the mask enables the exception
const
  WantedEx = [exDenormalized, exPrecision];
begin
  FSource := '';
  SetParserDecimalChar('.');
  SetParserListChar(',');
  ExternalTargets.Clear;
  oldExceptionMask := GetExceptionMask;
  SetExceptionMask(oldExceptionMask * WantedEx);
end;

procedure TTestParserHarness.TearDown;
begin
  FSource := '';
  SetExceptionMask(oldExceptionMask);
end;

procedure TTestParserHarness.TestOk(const value: string; wanted, eps: float; const msg: string);
var
  res: float;
begin
  try
    if eps < 0 then begin
      res := EvaluateExpression(value);
      if abs((res-wanted)/wanted) > -eps then
        CheckTrue(false, Format('Expected %g got %g %s', [wanted, res, msg]))
      else
        CheckTrue(true, 'so DUnit does not complain about missing test');
    end
    else
      CheckEquals(wanted, EvaluateExpression(value), eps, msg);
  except
    On E: EMathParserErr do begin
      CheckTrue(false, ParseMessage(E.ParserError));
    end;
  end;
end;

procedure TTestParserHarness.TestBadParse(const value: string; error: TParserError);
begin
  try
    EvaluateExpression(value);
  except
    On E: EMathParserErr do begin
      CheckEquals(ParseMessage(error), ParseMessage(E.ParserError));
    end;
  end;
end;

procedure TTestParserHarness.TestBadEval(const value: string; error: ExceptClass);
begin
  try
    EvaluateExpression(value);
  except
    On E: Exception do begin
      CheckEquals(error.ClassName, E.ClassName);
    end;
  end;
end;

procedure TTestTreeParser.TestOr;
begin
  TestOk('0 | 0', 0);
  TestOk('1 | 0', 1);
  TestOk('0 | 1', 1);
  TestOk('1 | 1', 1);
end;

procedure TTestTreeParser.TestAnd;
begin
  TestOk('0 & 0', 0);
  TestOk('1 & 0', 0);
  TestOk('0 & 1', 0);
  TestOk('1 & 1', 1);
end;

procedure TTestTreeParser.TestOverflow;
begin
  if sizeof(float) = sizeof(extended) then
    TestBadEval('1.1E4930 * 1.1E4930', EOverflow)
  else if sizeof(float) = sizeof(double) then
    TestBadEval('1.1E307 * 1.1E307', EOverflow)
  else if sizeof(float) = sizeof(single) then
    TestBadEval('1.1E38 * 1.1E38', EOverflow)
  else
    CheckTrue(false, 'Unknown float type');
end;

procedure TTestTreeParser.TestPower;
begin
  TestOk('2^3', power(2, 3));
end;

procedure TTestTreeParser.TestPower2;
begin
{$IFDEF EXPONENT_RIGHT_ASSOCIATIVE}
  {$IFDEF USE_MATH_UNIT}
  TestOk('2^3^4', power(2, power(3, 4)));
  {$ELSE}
  TestOk('2^3^4', power(2, power(3, 4)), -1E-16);
  {$ENDIF}
{$ELSE}
  {$IFDEF USE_MATH_UNIT}
  TestOk('2^3^4', power(power(2, 3), 4));
  {$ELSE}
  TestOk('2^3^4', power(power(2, 3), 4), -1E-16);
  {$ENDIF}
{$ENDIF}
end;

procedure TTestTreeParser.TestPower3;
begin
  TestOk('5^0', power(5, 0));
  TestOk('5^0', 1);
end;

procedure TTestTreeParser.TestPower4;
begin
  TestOk('0^0', power(0, 0));
  TestOk('0^0', 1);
end;

procedure TTestTreeParser.TestUnderflow;
begin
  if sizeof(float) = sizeof(extended) then
    TestBadEval('2/1E4930/1E4930' , EUnderflow)
  else if sizeof(float) = sizeof(double) then
    TestBadEval('2/1E306/1E306' , EUnderflow)
  else if sizeof(float) = sizeof(single) then
    TestBadEval('2/1E38/1E38' , EUnderflow)
  else
    CheckTrue(false, 'Unknown float type');
end;

procedure TTestTreeParser.TestCreate;
begin
  TestBadParse('',  peUnexpectedEOS);
end;

procedure TTestTreeParser.TestDivBy0;
begin
  TestBadEval('5/(2 - 2)', EZeroDivide);
end;

procedure TTestTreeParser.TestUnexpectedEOS;
begin
  TestBadParse('5 +', peUnexpectedEOS);
  TestBadParse('6 *', peUnexpectedEOS);
end;

procedure TTestTreeParser.TestWrongParamCount;
begin
  TestBadParse('sin(2, 8)', peTooManyParams);
  TestBadParse('log(2)', peTooFewParams);

end;

procedure TTestTreeParser.TestEqual;
begin
  TestOk('5 = 5', 1);
  TestOk('6 == 6', 1);
  TestOk('5 = 6', 0);
  TestOk('5 == 6', 0);
end;

procedure TTestTreeParser.TestExpectedEOS;
begin
  TestBadParse('5 + 8 32', peExpectedEOS);
  TestBadParse('6 * 8 32', peExpectedEOS);
  TestBadParse('7 * 8 !', peExpectedEOS);
  TestBadParse('8#', peExpectedEOS);
end;

procedure TTestTreeParser.TestUnexpectedElement;
begin
  TestBadParse('5 + \', peUnExpectedElement);
  TestBadParse('+*', peUnExpectedElement);
  TestBadParse('pi*sin(5)/sin(2+)', peUnexpectedElement);
  TestBadParse('sin(2+)', peUnexpectedElement);
end;

procedure TTestTreeParser.TestExpecteRightPar;
begin
  TestBadParse('5*(2 + 3', peExpectedRightPar);
end;

procedure TTestTreeParser.TestIf;
begin
  TestOk('if(1, 6, 5/0)', 6.0);
  TestOk('if(0, 6/0, 5)', 5.0);
end;


procedure TTestTreeParser.TestIntDiv;
begin
  TestOk('9 : 2', 9 div 2);
  TestOK('-9 : 2', (-9) div 2);
  TestOK('(-9) : 2', (-9) div 2);
  TestOK('9 : -2', 9 div -2);
  TestOK('-9 : -2', -9 div -2);
end;

procedure TTestTreeParser.TestMod;
begin
  TestOk('9 % 2', 9 mod 2);
  TestOK('-9 % 2', (-9) mod 2);
  TestOK('(-9) % 2', (-9) mod 2);
  TestOK('9 % -2', 9 mod -2);
  TestOK('-9 % -2', -9 mod -2);
end;

procedure TTestTreeParser.TestInvalidNumber;
begin
  TestBadParse('5.3E*8', peInvalidNumber);
  TestBadParse('5.3E3456789', peInvalidNumber);
  TestBadParse('5.3E-3456789', peInvalidNumber);
end;

procedure TTestTreeParser.TestLess;
begin
  TestOk('6 < 8', 1);
  TestOk('8 < 6', 0);
end;

procedure TTestTreeParser.TestLessOrEqual;
begin
  TestOk('6 <= 3 + 3', 1);
  TestOk('8 <= 6', 0);
end;


procedure TTestTreeParser.TestGreater;
begin
  TestOk('8 > 6', 1);
  TestOk('6 > 8', 0);
end;

procedure TTestTreeParser.TestGreaterOrEqual;
begin
  TestOk('6 >= 3 + 3', 1);
  TestOk('6 >= 8', 0);
end;



procedure TTestTreeParser.TestExpression1;
begin
  TestOk('5 - 32', 5 - 32);
end;

procedure TTestTreeParser.TestExpression2;
begin
  TestOk('5 - 32 + - 34', 5 - 32 + - 34);
end;

procedure TTestTreeParser.TestExpression3;
begin
  TestOk('5 - 32 * 4', 5 - 32 * 4);
end;

procedure TTestTreeParser.TestExpression4;
begin
  TestOk('5 - 32 * - 34', 5 - 32 * - 34);
end;

procedure TTestTreeParser.TestExpression5;
begin
  TestOk('(5 - 32) * 4', (5 - 32) * 4);
end;

procedure TTestTreeParser.TestExpression6;
begin
  TestOk('(6 - 2) / (3 - 1)', (6 - 2) / (3 - 1));
end;

procedure TTestTreeParser.TestFactor0;
begin
  TestOk('0', 0);
end;

procedure TTestTreeParser.TestFactor10;
begin
  TestOk(' 10 ', 10);
end;

procedure TTestTreeParser.TestFactorFloat;
begin
  TestOk(' 10.345E12', 10.345E12)
end;

procedure TTestTreeParser.TestFactorMinus10;
begin
  TestOk(' -  10 ', -10);
end;

procedure TTestTreeParser.TestFunctions;
begin
  TestOk('sin(pi/2)', sin(pi/2));
end;

procedure TTestTreeParser.TestMultiline;
begin
  TestOk(' 4 *'#13#10' 32 - '#13#10'5', 4 * 32 - 5);
end;

procedure TTestTreeParser.TestMultiline2;
begin
  TestOk(' 4 *'#10' 32 - '#10'5', 4 * 32 - 5);
end;

procedure TTestTreeParser.TestNot;
begin
  TestOk('!0', 1);
  TestOk('!1', 0);
  TestOk('!2', 0);
end;

procedure TTestTreeParser.TestNotEqual;
begin
  TestOk('5 != 6', 1);
  TestOk('6 <> 5', 1);
  TestOk('5 != 5', 0);
  TestOk('6 <> 6', 0);

  TestOk('0 != 0', 0);
  TestOk('0 != 1', 1);
  TestOk('1 != 0', 1);
  TestOk('1 != 1', 0);

end;


procedure TTestTreeParser.TestTerm1;
begin
  testOk('32.8 *72.22', 32.8 *72.22);
end;

procedure TTestTreeParser.TestTerm2;
begin
  testOk('32 * 72 / 8', 32 * 72 / 8);
end;

{ TTestTargets }


procedure TTestTargets.TestAddConstant;
begin
  CheckFalse(InternalTargets.HasTarget('halfPi'), 'Can''t do test "halfPi" already added');
  InternalTargets.Add('halfPi', TConstant.create(pi/2));
  testOk('halfPi', pi/2);
  InternalTargets.Delete('halfPi');
  CheckFalse(InternalTargets.HasTarget('halfPi'), '"HalfPi" found after deletion');
  testBadParse('halfPi', peUnknownIdentifier);
end;

type
  _sinx = class(TFunction)
  public
    function Calc: float; override;
  end;

function _sinx.Calc: float;
begin
  result := args[0]*sin(args[0]);
end;

procedure TTestTargets.TestAddFixedFunction;
begin
  CheckFalse(InternalTargets.HasTarget('sinx'), 'Can''t do test "sinx" already added');
  InternalTargets.Add('sinx', _sinx.create(1));
  testOk('sinx(2)', 2*sin(2));
  CheckTrue(InternalTargets.Delete('sinx'), '"sinx" not deleted');
  CheckFalse(InternalTargets.HasTarget('sinx'), '"sinx" found after deletion');
  testBadParse('3+sinx(9)', peUnknownIdentifier);
end;


var
  x: float;

procedure TTestTargets.TestAddVariable;
begin
  CheckFalse(InternalTargets.HasTarget('x'), 'Can''t do test "x" already added');
  InternalTargets.Add('x', TVariable.Create(@x));
  x := 3;
  testOk('2*x', 6);
  x := 6;
  testOk('x/2', 3);
  CheckTrue(InternalTargets.Delete('x'), '"x" not deleted');
  CheckFalse(InternalTargets.HasTarget('x'), '"x" found after deletion');
  testBadParse('2*x', peUnknownIdentifier);
end;

type
  _avg = class(TVariableFunction)
  public
    function Calc: float; override;
  end;

function _avg.Calc: float;
var
  i: integer;
begin
  result := 0;
  for i := 0 to Count - 1 do
    result := result + args[i]/count;
end;

procedure TTestTargets.TestAddVariableFunction;
begin
  InternalTargets.Add('TestAvg', _avg.create(1));
  testOk('TestAvg(3, 4, 5)', 4);
  InternalTargets.Delete('TestAvg');
  testBadParse('TestAvg(3,4)', peUnknownIdentifier);
end;

initialization
  RegisterTest(TTestTreeParser.Suite);
  RegisterTest(TTestTargets.Suite);
end.
